using System.Security.Cryptography.X509Certificates;
using AutoMapper;
using SoapEndpoint.Service.Proxy.Infrastructure;
using SoapEndpoint.Service.Proxy.Models.Flows;
using SoapEndpoint.Service.Proxy.Models.Systeem;
using SoapEndpoint.Service.Proxy.ResourceItems;
using SoapEndpoint.Tester.Infrastructure;
using SoapEndpoint.Tester.Properties;
using SoapEndpoint.Tester.Scenario;
using SoapEndpoint.Tester.Scenario.Implementations.Schuldbemiddeling;
using static SoapEndpoint.Tester.Infrastructure.FormHelperMethods;
using OrganisatieTypeV1 = Contracts.Generated.Flows.Schuldbemiddeling.v1.OrganisatieType;
using OrganisatieTypeV2 = Contracts.Generated.Systeem.v2.OrganisatieType;
using SchuldbemiddelingMessageType = SoapEndpoint.Service.Systeem.v1.LogbookService.Berichttype;
using SbEnSkMessageType = SoapEndpoint.Service.Flows.SbEnSk.v1.LogbookService.BerichtType;
using WaarborgfondsMessageType = SoapEndpoint.Service.Flows.Waarborgfonds.v1.LogbookService.BerichtType;
using InventarisatieMessageType = SoapEndpoint.Service.Flows.Inventarisatie.v1.LogbookService.BerichtType;
using BrEnHfMessageType = SoapEndpoint.Service.Flows.BrEnHf.v1.LogbookService.BerichtType;
using BewindvoeringMessageType = SoapEndpoint.Service.Flows.Bewindvoering.v1.LogbookService.BerichtType;
using Environment = SoapEndpoint.Tester.Infrastructure.Environment;
using PauzeSignaal3MessageType = SoapEndpoint.Service.Flows.PauzeSignaal3.v1.LogbookService.BerichtType;
using ShvSignaalMessageType = SoapEndpoint.Service.Flows.ShvSignaal.v1.LogbookService.BerichtType;

namespace SoapEndpoint.Tester;

public partial class Form1 : Form
{
    private const FlowType DefaultFlowType = FlowType.Schuldbemiddeling;

    private readonly IDictionary<FlowType, ListBox> _supportedFlowsBoxes = new Dictionary<FlowType, ListBox>();
    private ServiceUrlInfo _urlInfo;

    public Form1()
    {
        InitializeComponent();
    }

    private ResourceItemManager ResourceItemManager { get; } = new();
    private IList<CertificateItem> Certificates { get; set; }
    public Environment CurrentEnvironment => cbCurrentEnvironment.SelectedItem as Environment;
    public Version CurrentSystemVersion => cbCurrentSystemVersion.SelectedItem as Version;
    public Version CurrentShvFlowVersion => cbShvMessageVersion.SelectedItem as Version;
    public Version CurrentSeFlowVersion => cbSeMessageVersion.SelectedItem as Version;
    public Version CurrentConversationFlowVersion => cbConversationFlowVersion.SelectedItem as Version;
    public Version CurrentLogbookFlowVersion => cbLogbookFlowVersion.SelectedItem as Version;

    public FlowType? SelectedShvFlowType { get; set; }
    public FlowType? SelectedSeFlowType { get; set; }
    public FlowType? SelectedScenarioFlowType { get; set; }
    public FlowType? SelectedConversationFlowType { get; set; }
    public FlowType? SelectedLogbookFlowType { get; set; }

    public ResourceItem<SchuldbemiddelingMessageType> SelectedShvSchuldbemiddelingResourceItem { get; set; }
    public ResourceItem<SbEnSkMessageType> SelectedShvSbEnSkResourceItem { get; set; }
    public ResourceItem<WaarborgfondsMessageType> SelectedShvWaarborgfondsResourceItem { get; set; }
    public ResourceItem<InventarisatieMessageType> SelectedShvInventarisatieResourceItem { get; set; }
    public ResourceItem<BrEnHfMessageType> SelectedShvBrEnHfResourceItem { get; set; }
    public ResourceItem<BewindvoeringMessageType> SelectedShvBewindvoeringResourceItem { get; set; }
    public ResourceItem<PauzeSignaal3MessageType> SelectedShvPauzeSignaal3ResourceItem { get; set; }
    public ResourceItem<ShvSignaalMessageType> SelectedShvSignaalResourceItem { get; set; }
    public ResourceItem<SchuldbemiddelingMessageType> SelectedSeSchuldbemiddelingResourceItem { get; set; }
    public ResourceItem<SbEnSkMessageType> SelectedSeSbEnSkResourceItem { get; set; }
    public ResourceItem<WaarborgfondsMessageType> SelectedSeWaarborgfondsResourceItem { get; set; }
    public ResourceItem<InventarisatieMessageType> SelectedSeInventarisatieResourceItem { get; set; }
    public ResourceItem<BrEnHfMessageType> SelectedSeBrEnHfResourceItem { get; set; }
    public ResourceItem<BewindvoeringMessageType> SelectedSeBewindvoeringResourceItem { get; set; }
    public ResourceItem<PauzeSignaal3MessageType> SelectedSePauzeSignaal3ResourceItem { get; set; }

    public IList<CertificateItem> ShvCertificates { get; set; }
    public IList<CertificateItem> SeCertificates { get; set; }

    public ScenarioManager AutoReplyManager { get; set; }

    public IMapper Mapper { get; set; }

    public string ShvContent => rtbShvMessageContent.Text;
    public string SeContent => rtbSeContent.Text;

    public string LogKey => TxtLogKey.Text;
    private List<BijlageModel> ShvAddedAttachments { get; } = new();
    private List<BijlageModel> SeAddedAttachments { get; } = new();

    private IList<ScenarioResultGroup> ScenarioResultGroups { get; set; }

    private void Form1_Load(object sender, EventArgs e)
    {
        LblConversationState.Text = string.Empty;

        TxtLogKey.Text = Program.Configuration["LogKey"];

        BindCertificates();

        SetDefaultValues();

        SetGuid(txtConversationRef);
        SetGuid(txtReference);
        SetGuid(txtSeReference);

        HighlightColors();

        dgvScenarioResult.SelectionChanged += DgvScenarioResult_SelectionChanged;
        dgvScenarioConverstations.CellContentClick += DgvScenarioConversations_CellContentClick;

        InitializeEnvironments();

        BindOrganizationTypes(cbOrganisationType, CurrentSystemVersion);
        BindOrganizationTypes(cbLogSenderOrganizationType, CurrentSystemVersion);
        BindOrganizationTypes(cbLogReceiverOrganizationType, CurrentSystemVersion);

        Mapper = AutoMapperHelper.InitializeAutoMapper();

        _urlInfo = new ServiceUrlInfo(Mapper);

        tabControl1.TabPages.Remove(AutoReply);
    }

    private void BindCertificates()
    {
        var loadAllCertificatesResult = bool.TryParse(Program.Configuration["LoadAllCertificates"],
            out var loadAllCertificates);

        if (loadAllCertificatesResult && loadAllCertificates)
        {
            // Load all certificates
            Certificates = LoadCertificates();
            ShvCertificates = Certificates.ToList();
            SeCertificates = Certificates.ToList();
        }
        else
        {
            // Only load NVVK self singed certificates
            using (var rootCertificate = new X509Certificate2(Resources.Nvvk_Root_Cert))
            {
                Certificates = LoadCertificates(rootCertificate);
            }

            ShvCertificates = Certificates.Where(c => c.Type == OrganisatieType.Schuldhulpverlener).ToList();
            SeCertificates = Certificates.Where(c => c.Type == OrganisatieType.Schuldeiser).ToList();
        }

        BindCertificate(cbShvCertificate, ShvCertificates);
        BindCertificate(cbSeCertificate, SeCertificates);

        BindCertificate(cbLogCertificate, Certificates);
        BindCertificate(cbMessageBoxCertificate, Certificates);
        BindCertificate(cbAddresbookCertificate, Certificates);
        BindCertificate(cbOrganizationCertficate, Certificates);
        BindCertificate(cbConversationCertificate, Certificates);
        BindCertificate(cbScenarioCertificate, Certificates);
        BindCertificate(cbAutoReplyCertificate1, ShvCertificates, "No SHV certificate");
        BindCertificate(cbAutoReplyCertificate2, SeCertificates, "No SE certificate");
    }

    private void InitializeEnvironments()
    {
        Environments.InitializeEnvironments();

        var defaultVersion = SystemVersions.DefaultVersion(Program.Configuration["DefaultVersion"]);

        BindEnvironments(cbCurrentEnvironment, Environments.DefaultEnvironment);
        BindVersions(cbCurrentSystemVersion, SystemVersions.MajorVersions, defaultVersion);
        BindFlows(cbShvFlowType, DefaultFlowType);
        BindFlows(cbSeFlowType, DefaultFlowType);
        BindFlows(cbScenarioFlowType, DefaultFlowType);
        BindFlows(cbConversationFlowType, DefaultFlowType);
        BindFlows(cbLogbookFlowType, DefaultFlowType);
    }

    private async Task LoadOrganisationsAsync()
    {
        AddMessage("Loading organizations...");

        int? ConvertOrganisatieType(string type, Version systemVersion)
        {
            switch (systemVersion.Major)
            {
                case 1:
                {
                    var parsed = Enum.TryParse(type, out OrganisatieTypeV1 result);
                    return parsed ? (int?)result : null;
                }
                case 2:
                {
                    var parsed = Enum.TryParse(type, out OrganisatieTypeV2 result);
                    return parsed ? (int?)result : null;
                }
                default:
                    return null;
            }
        }

        await ExecuteAsync(async () =>
        {
            var proxy = _urlInfo.GetAddressProxy(GetCertificate(cbAddresbookCertificate), CurrentEnvironment,
                CurrentSystemVersion);
            if (proxy == null)
            {
                AddMessage("Something went wrong with creating the proxy.", false);
                return;
            }

            var message = new AdresboekZoekBerichtModel
            {
                Naam = txtOrgNameFilter.Text,
                OrganisatieType =
                    ConvertOrganisatieType(cbOrganisationType.SelectedItem.ToString(), CurrentSystemVersion)
            };

            if (cbIsNvvkMember.CheckState != CheckState.Indeterminate)
            {
                message.IsNvvkLid = cbIsNvvkMember.Checked;
            }

            if (cbCollectiveDebtSettlement.CheckState != CheckState.Indeterminate)
            {
                message.CollectiefSchuldRegelen = cbCollectiveDebtSettlement.Checked;
            }

            if (cbCanAgreeInAdvance.CheckState != CheckState.Indeterminate)
            {
                message.BijVoorbaatAkkoord = cbCanAgreeInAdvance.Checked;
            }

            var response = await proxy.GetOrganizationsAsync(message);

            AddMessage($"Loaded {response.Length} organisations into address book.");

            BindAddressGrid(response);
        });
    }

    private void HighlightColors()
    {
        RtbHelper.HighlightRTF(rtbShvMessageContent);
        RtbHelper.HighlightRTF(rtbSeContent);
        RtbHelper.HighlightRTF(rtbMessageBoxMessageContent);
    }

    private void SetDefaultCertificate(string thumbprint, ComboBox combobox)
    {
        if (!string.IsNullOrWhiteSpace(thumbprint))
        {
            var certificate = Certificates.FirstOrDefault(c =>
                string.Equals(c.Certificate.Thumbprint, thumbprint, StringComparison.OrdinalIgnoreCase));

            if (certificate != null)
            {
                combobox.SelectedItem = certificate;
            }
        }
    }

    private void SetDefaultValues()
    {
        BindVersions(ListSchuldbemiddelingVersions, FlowDefinitions.SchuldBemiddelingFlow.SupportedVersions);
        BindVersions(ListSbEnSkVersions, FlowDefinitions.SbEnSkFlow.SupportedVersions);
        BindVersions(ListWaarborgfondsVersions, FlowDefinitions.WaarborgfondsFlow.SupportedVersions);
        BindVersions(ListInventarisatieVersions, FlowDefinitions.InventarisatieFlow.SupportedVersions);
        BindVersions(ListBrEnHfVersions, FlowDefinitions.BrEnHfFlow.SupportedVersions);
        BindVersions(ListBewindvoeringVersions, FlowDefinitions.BewindvoeringFlow.SupportedVersions);
        BindVersions(ListPauzeSignaal3Versions, FlowDefinitions.PauzeSignaal3Flow.SupportedVersions);
        BindVersions(ListShvSignaalVersions, FlowDefinitions.ShvSignaalFlow.SupportedVersions);

        _supportedFlowsBoxes.Add(FlowType.Schuldbemiddeling, ListSchuldbemiddelingVersions);
        _supportedFlowsBoxes.Add(FlowType.SbEnSk, ListSbEnSkVersions);
        _supportedFlowsBoxes.Add(FlowType.Waarborgfonds, ListWaarborgfondsVersions);
        _supportedFlowsBoxes.Add(FlowType.Inventarisatie, ListInventarisatieVersions);
        _supportedFlowsBoxes.Add(FlowType.BrEnHf, ListBrEnHfVersions);
        _supportedFlowsBoxes.Add(FlowType.Bewindvoering, ListBewindvoeringVersions);
        _supportedFlowsBoxes.Add(FlowType.PauzeSignaal3, ListPauzeSignaal3Versions);
        _supportedFlowsBoxes.Add(FlowType.ShvSignaal, ListShvSignaalVersions);

        SetDefaultCertificate(Program.Configuration["ShvCertificateThumbprint"], cbShvCertificate);
        SetDefaultCertificate(Program.Configuration["SeCertificateThumbprint"], cbSeCertificate);
        txtShvReceiverReference.Text = Program.Configuration["ReceiverReference"];
    }

    private async Task ExecuteShvActionAsync<TMessageType>(ResourceItem<TMessageType> item)
        where TMessageType : Enum
    {
        var header = new BerichtHeaderModel
        {
            ConversatieReferentienummer = ParseOptionalGuid(txtConversationRef.Text),
            CrossReferentienummer = ParseOptionalGuid(txtCrossReference.Text),
            Ontvanger = ParseOptionalGuid(txtShvReceiverReference.Text),
            Referentienummer = Guid.Parse(txtReference.Text),
            Bijlages = ShvAddedAttachments,
            Versie = item.ResourceManagerModel.VersionAsString
        };

        async Task<ShvKnooppuntBerichtBevestigingModel> ProxyCallAsync()
        {
            var certificate = GetCertificate(cbShvCertificate);
            var message = new BerichtModel<TMessageType>(ShvContent, item.TypeName);

            var proxy = _urlInfo
                .GetShvProxy<TMessageType>(certificate, SelectedShvFlowType ?? DefaultFlowType, CurrentEnvironment,
                    CurrentShvFlowVersion);
            return await proxy.SendMessageAsync(header, message);
        }

        await MessageHelper.ExecActionAsync(cbXmlValidation.Checked, ShvContent, header, item,
            ProxyCallAsync, AddMessage, CurrentShvFlowVersion.ToString());

        ShvAddedAttachments.Clear();
    }

    private async Task ExecuteSeActionAsync<TMessageType>(ResourceItem<TMessageType> item,
        BerichtHeaderModel header = null)
        where TMessageType : Enum
    {
        header ??= new BerichtHeaderModel
        {
            Referentienummer = ParseGuid(txtSeReference.Text),
            CrossReferentienummer = ParseOptionalGuid(txtSeCrossReference.Text),
            ConversatieReferentienummer = ParseOptionalGuid(txtSeConversationReference.Text),
            Versie = item.ResourceManagerModel.VersionAsString
        };

        async Task<ShvKnooppuntBerichtBevestigingModel> ProxyCallAsync()
        {
            var certificate = GetCertificate(cbSeCertificate);
            var message = new BerichtModel<TMessageType>(SeContent, item.TypeName);

            return await _urlInfo
                .GetSeProxy<TMessageType>(certificate, SelectedSeFlowType ?? DefaultFlowType, CurrentEnvironment,
                    CurrentSeFlowVersion).SendMessageAsync(header, message);
        }

        await MessageHelper.ExecActionAsync(cbXmlValidation.Checked, SeContent, header, item,
            ProxyCallAsync, AddMessage, CurrentSeFlowVersion.ToString());

        SeAddedAttachments.Clear();
    }

    private BerichtHeaderModel GetBerichtHeaderWaarborgfonds(WaarborgfondsMessageType messageType)
    {
        var waarborgfondsInitiatingMessageTypes = new[]
        {
            WaarborgfondsMessageType.SE_SaneringskredietInformatieAanvraag,
            WaarborgfondsMessageType.SE_BorgstellingIngetrokken
        };

        //When sending a gebeurtenis, the waarborgfonds replies to an existing conversation,
        //so we do not need to send a receiver reference as that will be retrieved from the conversation entity
        var waarborgfondsGebeurtenisMessageHeader = new BerichtHeaderModel
        {
            Referentienummer = ParseGuid(txtSeReference.Text),
            CrossReferentienummer = ParseOptionalGuid(txtSeCrossReference.Text),
            Versie = CurrentSeFlowVersion.ToString(),
            Bijlages = SeAddedAttachments
        };

        //When sending a waarborgfonds message that is an initiating message(i.e starts a conversation)
        //we do not need to send a conversation reference. Instead, we send the receiver reference.
        var waarborgfondsInitiatingMessageHeader =
            new BerichtHeaderModel
            {
                Ontvanger = ParseOptionalGuid(txtSeReceiverReference.Text),
                Referentienummer = ParseGuid(txtSeReference.Text),
                Versie = CurrentSeFlowVersion.ToString(),
                Bijlages = SeAddedAttachments
            };

        return waarborgfondsInitiatingMessageTypes.Contains(messageType)
            ? waarborgfondsInitiatingMessageHeader
            : waarborgfondsGebeurtenisMessageHeader;
    }

    private async Task<bool> ExecuteAsync(Func<Task> action, string message = null, string errorMessage = null)
    {
        return await ExecuteActionAsync(action,
            () => AddMessage(message),
            ex => AddMessage(errorMessage + " " + ex.Message, false));
    }

    public IList<CertificateItem> LoadCertificates(X509Certificate2 rootCertificate = null)
    {
        var certificates = default(IList<CertificateItem>);
        try
        {
            certificates = CertificateHelper.FindCertificates(rootCertificate).OrderBy(x => x.Name).ToList();

            if (certificates.Any() != true)
            {
                AddMessage("No certificates were found.");
            }
        }
        catch (Exception)
        {
            AddMessage("Failed to load the certificates, try running as administrator");
        }

        return certificates;
    }

    private async Task ConfirmMessagesAsync(X509Certificate2 certificate, string guidStr)
    {
        await ExecuteAsync(async () =>
        {
            var proxy = _urlInfo.GetMessageBoxProxy(certificate, CurrentEnvironment, CurrentSystemVersion);
            if (proxy == null)
            {
                AddMessage("Something went wrong with creating the proxy.", false);
                return;
            }

            if (Guid.TryParse(guidStr, out var guid))
            {
                var messages = await proxy.ConfirmBatchAsync(new MessageboxConfirmatieBerichtModel(guid));
                AddMessage($"Messages confirmed: {messages.Aantal}");
            }

            dgvMessageBoxMessages.Rows.Clear();
            txtMessageBoxBatchId.Clear();
        });
    }

    private async Task GetMessagesAsync(X509Certificate2 certificate)
    {
        MessageboxBerichtinformatieModel messages;
        await ExecuteAsync(async () =>
        {
            var proxy = _urlInfo.GetMessageBoxProxy(certificate, CurrentEnvironment, CurrentSystemVersion);
            if (proxy == null)
            {
                AddMessage("Something went wrong with creating the proxy.", false);
                return;
            }

            messages = await proxy.GetMessagesAsync();

            txtMessageBoxBatchId.Clear();
            rtbMessageBoxMessageContent.Clear();

            if (messages != null)
            {
                lblMessageBoxBatchValidTo.Text =
                    $@"Created on: {messages.AangemaaktOp} Valid to: {messages.GeldigTot}" +
                    $@" Batch messages {messages.AantalBerichtenInBatch} New messages: {messages.AantalNieuweBerichtenBuitenBatch}";

                txtMessageBoxBatchId.Text = messages.BatchReferentie.ToString();
                BindData(dgvMessageBoxMessages, messages.Berichten, GetMessageBoxRow);
            }

            AddMessage($"Messages retrieved: {messages?.Berichten?.Count ?? 0}");
        });
    }

    private void BindAddressGrid(IList<AdresboekOrganisatieModel> response)
    {
        BindData(dataGridView2, response, i => new object[]
        {
            i.Referentienummer,
            i.Naam,
            i.OrganisatieType,
            i.Sector,
            i.KvkNummer,
            i.NvvkLid,
            i.CollectiefSchuldRegelen,
            i.BijVoorbaatAkkoord,
            BindDataGridFlows(i.OndersteundeFlows),
            i.Geblokkeerd,
            i.RedenGeblokkeerd
        });
    }

    private string BindDataGridFlows(List<OrganizationFlowModel> ondersteundeFlows)
    {
        return string.Concat(ondersteundeFlows.ConvertAll(flow =>
            $"{flow.OrganisatieFlowType} - {string.Join(", ", flow.OndersteundeVersies)}" + "\n"));
    }

    private async Task LoadLogbookAsync<TMessageType>() where TMessageType : struct, Enum
    {
        int? ConvertOrganisatieType(string type, Version systemVersion)
        {
            switch (systemVersion.Major)
            {
                case 1:
                {
                    var parsed = Enum.TryParse(type, out OrganisatieTypeV1 result);
                    return parsed ? (int?)result : null;
                }
                case 2:
                {
                    var parsed = Enum.TryParse(type, out OrganisatieTypeV2 result);
                    return parsed ? (int?)result : null;
                }
                default:
                    return null;
            }
        }

        await ExecuteAsync(async () =>
        {
            var certificate = GetCertificate(cbLogCertificate);
            var proxy = _urlInfo.GetLogbookProxy<TMessageType>(certificate, CurrentEnvironment,
                CurrentSystemVersion, CurrentLogbookFlowVersion, SelectedLogbookFlowType ?? DefaultFlowType);
            if (proxy == null)
            {
                AddMessage("Something went wrong with creating the proxy.", false);
                return;
            }

            Guid? ParseNullableGuid(TextBox txt)
            {
                return Guid.TryParse(txt.Text, out var result) ? result : default(Guid?);
            }

            var message = new LogboekZoekberichtModel
            {
                Referentienummer = ParseNullableGuid(txtLogReference),
                ConversatieReferentienummer = ParseNullableGuid(txtLogConversationReference),
                CrossReferentienummer = ParseNullableGuid(txtLogCrossReference),
                Afzender = ParseNullableGuid(txtLogSender),
                Ontvanger = ParseNullableGuid(txtLogReceiver),
                Versie = CurrentLogbookFlowVersion.ToString(),
                BerichtType = Enum.TryParse<TMessageType>(cbLogMessageType.Text, out var berichtType)
                    ? berichtType.ToString()
                    : null,
                AfzenderOrganisatieType =
                    ConvertOrganisatieType(cbLogSenderOrganizationType.Text, CurrentSystemVersion),
                OntvangerOrganisatieType =
                    ConvertOrganisatieType(cbLogReceiverOrganizationType.Text, CurrentSystemVersion)
            };

            if (cbLogRetrieved.CheckState != CheckState.Indeterminate)
            {
                message.Opgehaald = cbLogRetrieved.Checked;
            }

            if (cbLogConfirmed.CheckState != CheckState.Indeterminate)
            {
                message.Bevestigd = cbLogConfirmed.Checked;
            }

            if (dtLogFrom.Checked)
            {
                message.VanafDatum = new DateTime(dtLogFrom.Value.Year, dtLogFrom.Value.Month, dtLogFrom.Value.Day, 0,
                    0, 0);
            }

            if (dtLogTo.Checked)
            {
                message.TotDatum = new DateTime(dtLogFrom.Value.Year, dtLogFrom.Value.Month, dtLogFrom.Value.Day, 23,
                    59, 59);
            }

            var response = await proxy.GetMessagesAsync(message);

            BindData(logBookDataGridView, response, i => new object[]
            {
                i.Referentienummer,
                i.ConversatieReferentienummer,
                i.CrossReferentienummer,
                i.Afzender,
                i.Ontvanger,
                i.BerichtType,
                i.Opgehaald,
                i.Bevestigd,
                i.Datum,
                i.Versie
            });
        });
    }

    private Task LoadOrganizationDetailAsync()
    {
        return ExecuteAsync(async () =>
        {
            var certificate = GetCertificate(cbOrganizationCertficate);

            ListSchuldbemiddelingVersions.ClearSelected();
            ListSbEnSkVersions.ClearSelected();
            ListBrEnHfVersions.ClearSelected();
            ListInventarisatieVersions.ClearSelected();
            ListWaarborgfondsVersions.ClearSelected();
            ListBewindvoeringVersions.ClearSelected();
            ListPauzeSignaal3Versions.ClearSelected();

            var proxy = _urlInfo.GetAddressProxy(certificate, CurrentEnvironment, CurrentSystemVersion);
            if (proxy == null)
            {
                AddMessage("Something went wrong with creating the proxy.", false);
                return;
            }

            var result = await proxy.GetOrganizationAsync();

            if (result == null)
            {
                return;
            }

            txtOrgName.Text = result.Naam;
            txtOrgReference.Text = result.Referentienummer.ToString();
            txtOrgSector.Text = result.Sector;
            txtOrgMessageType.Text = result.BerichtSE.ToString();
            txtOrgKvkNumber.Text = result.KvkNummer;
            txtOrgEstablishmentNumber.Text = result.Vestigingsnummer;
            txtOrgApplicationName.Text = result.Applicatienaam;

            txtOrgStreet.Text = result.Adres.Straatnaam;
            txtOrgHouseNumber.Text = result.Adres.Huisnummer;
            txtOrgIndication.Text = result.Adres.AanduidingBijHuisnummer;
            txtOrgLetter.Text = result.Adres.HuisnummerLetter;
            txtOrgAddition.Text = result.Adres.HuisnummerToevoeging;
            txtOrgCity.Text = result.Adres.Woonplaats;
            txtOrgZipCode.Text = result.Adres.Postcode;
            txtOrgHouseBoatReference.Text = result.Adres.Woonbootverwijzing;
            txtOrgCaravanReference.Text = result.Adres.Woonwagenverwijzing;

            txtOrgContactName.Text = result.Contactpersoon.Naam;
            txtOrgContactEmail.Text = result.Contactpersoon.Emailadres;
            txtOrgContactPhoneNumber.Text = result.Contactpersoon.Telefoonnummer;

            OrganizationToestemmingsverklaringCheckbox.Checked = result.Toestemmingsverklaring == true;
            organizationVtlbCheckbox.Checked = result.Vtlb == true;

            foreach (var flow in result.OndersteundeFlows)
            {
                var versionStrings = _supportedFlowsBoxes[flow.OrganisatieFlowType].Items.Cast<Version>()
                    .Select(x => x.ToString()).ToArray();
                var indices = flow.OndersteundeVersies.Select(v => Array.IndexOf(versionStrings, v));
                foreach (var index in indices) _supportedFlowsBoxes[flow.OrganisatieFlowType].SetSelected(index, true);
            }
        });
    }

    private Task UpdateOrganizationAsync()
    {
        return ExecuteAsync(async () =>
        {
            var certificate = GetCertificate(cbOrganizationCertficate);
            var proxy = _urlInfo.GetAddressProxy(certificate, CurrentEnvironment, CurrentSystemVersion);
            if (proxy == null)
            {
                AddMessage("Something went wrong with creating the proxy.", false);
                return;
            }

            var bericht = new UpdateOrganisatieBerichtModel
            {
                Applicatienaam = !string.IsNullOrWhiteSpace(txtOrgApplicationName.Text)
                    ? txtOrgApplicationName.Text
                    : null,
                Adres = new AdresModel
                {
                    Straatnaam = !string.IsNullOrWhiteSpace(txtOrgStreet.Text) ? txtOrgStreet.Text : null,
                    Huisnummer = !string.IsNullOrWhiteSpace(txtOrgHouseNumber.Text) ? txtOrgHouseNumber.Text : null,
                    AanduidingBijHuisnummer = !string.IsNullOrWhiteSpace(txtOrgIndication.Text)
                        ? txtOrgIndication.Text
                        : null,
                    HuisnummerLetter = !string.IsNullOrWhiteSpace(txtOrgLetter.Text) ? txtOrgLetter.Text : null,
                    HuisnummerToevoeging =
                        !string.IsNullOrWhiteSpace(txtOrgAddition.Text) ? txtOrgAddition.Text : null,
                    Woonplaats = !string.IsNullOrWhiteSpace(txtOrgCity.Text) ? txtOrgCity.Text : null,
                    Postcode = !string.IsNullOrWhiteSpace(txtOrgZipCode.Text) ? txtOrgZipCode.Text : null,
                    Woonbootverwijzing = !string.IsNullOrWhiteSpace(txtOrgHouseBoatReference.Text)
                        ? txtOrgHouseBoatReference.Text
                        : null,
                    Woonwagenverwijzing = !string.IsNullOrWhiteSpace(txtOrgCaravanReference.Text)
                        ? txtOrgCaravanReference.Text
                        : null
                },
                Contactpersoon = new ContactpersoonModel
                {
                    Naam = !string.IsNullOrWhiteSpace(txtOrgContactName.Text) ? txtOrgContactName.Text : null,
                    Emailadres = !string.IsNullOrWhiteSpace(txtOrgContactEmail.Text)
                        ? txtOrgContactEmail.Text
                        : null,
                    Telefoonnummer = !string.IsNullOrWhiteSpace(txtOrgContactPhoneNumber.Text)
                        ? txtOrgContactPhoneNumber.Text
                        : null
                }
            };

            await proxy.UpdateOrganizationAsync(bericht);

            AddMessage("Organization updated");
        });
    }

    private async void ShowConversationDetail<TMessageType>(Guid conversationRef, X509Certificate2 certificate)
        where TMessageType : Enum
    {
        tabControl1.SelectedTab = Conversation;

        await ShowConversationAsync<TMessageType>(conversationRef, certificate);
    }

    private Task ShowConversationAsync<TMessageType>(Guid? reference, X509Certificate2 certificate)
        where TMessageType : Enum
    {
        return ExecuteAsync(async () =>
        {
            LblConversationState.ResetText();
            txtConversationRefState.Text = reference.ToString();
            cbConversationCertificate.SelectedItem =
                Certificates.FirstOrDefault(c => c.Certificate.Thumbprint == certificate.Thumbprint);

            var proxy = _urlInfo.GetLogbookProxy<TMessageType>(certificate, CurrentEnvironment,
                CurrentSystemVersion, CurrentConversationFlowVersion,
                SelectedConversationFlowType ?? DefaultFlowType);
            if (proxy == null)
            {
                AddMessage("Something went wrong with creating the proxy.", false);
                return;
            }

            var message = new ConversatieLogboekZoekberichtModel
            {
                Referentienummer = reference
            };

            var result = await proxy.GetConversationAsync(message);

            LblConversationState.Text = result.Staat;

            BindData(DgvConversation, result.Berichten, i => new object[]
            {
                i.Referentienummer,
                i.ConversatieReferentienummer,
                i.CrossReferentienummer,
                i.Afzender,
                i.Ontvanger,
                i.BerichtType,
                i.Opgehaald,
                i.Bevestigd,
                i.Datum,
                i.Versie
            });
        });
    }

    private async void BtnCheckConversationState_Click(object sender, EventArgs e)
    {
        switch (SelectedConversationFlowType)
        {
            case FlowType.Schuldbemiddeling:
                await ShowConversationAsync<SchuldbemiddelingMessageType>(
                    ParseOptionalGuid(txtConversationRefState.Text),
                    GetCertificate(cbConversationCertificate));
                break;
            case FlowType.SbEnSk:
                await ShowConversationAsync<SbEnSkMessageType>(
                    ParseOptionalGuid(txtConversationRefState.Text),
                    GetCertificate(cbConversationCertificate));
                break;
            case FlowType.Inventarisatie:
                await ShowConversationAsync<InventarisatieMessageType>(
                    ParseOptionalGuid(txtConversationRefState.Text),
                    GetCertificate(cbConversationCertificate));
                break;
            case FlowType.Waarborgfonds:
                await ShowConversationAsync<WaarborgfondsMessageType>(
                    ParseOptionalGuid(txtConversationRefState.Text),
                    GetCertificate(cbConversationCertificate));
                break;
            case FlowType.BrEnHf:
                await ShowConversationAsync<BrEnHfMessageType>(
                    ParseOptionalGuid(txtConversationRefState.Text),
                    GetCertificate(cbConversationCertificate));
                break;
            case FlowType.Bewindvoering:
                await ShowConversationAsync<BewindvoeringMessageType>(
                    ParseOptionalGuid(txtConversationRefState.Text),
                    GetCertificate(cbConversationCertificate));
                break;
            case FlowType.PauzeSignaal3:
                await ShowConversationAsync<PauzeSignaal3MessageType>(
                    ParseOptionalGuid(txtConversationRefState.Text),
                    GetCertificate(cbConversationCertificate));
                break;
            case FlowType.ShvSignaal:
                await ShowConversationAsync<ShvSignaalMessageType>(
                    ParseOptionalGuid(txtConversationRefState.Text),
                    GetCertificate(cbConversationCertificate));
                break;
            default:
                throw new ArgumentOutOfRangeException(
                    $"Conversation can't be checked for {SelectedConversationFlowType}");
        }
    }

    private void LoadScenarios(FlowType flowType)
    {
        CheckedLbScenarios.Items.Clear();
        var scenariosExist = Scenario.Scenarios.Items.TryGetValue(flowType, out var scenarios);
        if (!scenariosExist)
        {
            return;
        }

        foreach (var item in scenarios) CheckedLbScenarios.Items.Add(item, false);
    }

    private async void BtnRunScenarios_Click(object sender, EventArgs e)
    {
        try
        {
            ClearScenarioGridViews();

            ScenarioResultGroups = new List<ScenarioResultGroup>();

            var scenarios = CheckedLbScenarios.CheckedItems.Cast<ScenarioBase>().ToList();

            BtnRunScenarios.Enabled = false;

            //Get the address book.
            AddMessage("Fetching the address book");
            var addressBookCertificate = GetCertificate(cbScenarioCertificate);
            var proxy = _urlInfo.GetAddressProxy(addressBookCertificate, CurrentEnvironment, CurrentSystemVersion);
            if (proxy == null)
            {
                AddMessage("Something went wrong with creating the proxy.", false);
                return;
            }

            var organizations = await proxy.GetOrganizationsAsync(new AdresboekZoekBerichtModel());

            AddMessage($"Starting to execute {scenarios.Count} scenarios.");

            foreach (var scenario in scenarios)
                try
                {
                    var manager = new ScenarioManager(
                        scenario,
                        Certificates,
                        organizations,
                        ResourceItemManager,
                        _urlInfo,
                        AddMessage,
                        CurrentSystemVersion,
                        CurrentEnvironment);

                    var resultGroup = await manager.ExecuteScenarioAsync();
                    if (resultGroup != null)
                    {
                        ScenarioResultGroups.Add(resultGroup);
                    }
                    else
                    {
                        AddMessage($"Skipping scenario: {scenario.Name}", false);
                    }
                }
                catch (Exception scenarioEx)
                {
                    AddMessage($"Scenario failed: {scenarioEx.Message}", false);
                    throw;
                }

            BindData(dgvScenarioResult, ScenarioResultGroups, group => new object[]
            {
                group.Scenario.Name,
                group.Scenario.Shv?.Naam,
                group.Scenario.Se?.Naam,
                group.Success
            });

            if (dgvScenarioResult.Rows.Count > 0)
            {
                dgvScenarioResult.Rows[0].Selected = true;
                DgvScenarioResult_SelectionChanged(dgvScenarioResult, EventArgs.Empty);
            }
        }
        catch (Exception ex)
        {
            AddMessage($"Exception occured: {ex.Message}", false);
        }

        BtnRunScenarios.Enabled = true;
    }

    private void ClearScenarioGridViews()
    {
        if (dgvScenarioResult.RowCount > 0)
        {
            dgvScenarioResult.Rows.Clear();
        }

        if (dgvScenarioConverstations.RowCount > 0)
        {
            dgvScenarioConverstations.Rows.Clear();
        }
    }

    private void DgvScenarioConversations_CellContentClick(object sender, DataGridViewCellEventArgs e)
    {
        var dgv = sender as DataGridView;

        var result = dgv?.Rows[e.RowIndex].Tag as ScenarioResult;
        if (result?.Conversation == null)
        {
            dgv?.Rows.Clear();
            return;
        }

        switch (SelectedScenarioFlowType)
        {
            case FlowType.Schuldbemiddeling:
                ShowConversationDetail<SchuldbemiddelingMessageType>(result.Conversation.Reference,
                    result.Scenario.Manager.ShvCertificate);
                break;
            case FlowType.SbEnSk:
                ShowConversationDetail<SbEnSkMessageType>(result.Conversation.Reference,
                    result.Scenario.Manager.ShvCertificate);
                break;
            case FlowType.Waarborgfonds:
                ShowConversationDetail<WaarborgfondsMessageType>(result.Conversation.Reference,
                    result.Scenario.Manager.ShvCertificate);
                break;
            case FlowType.Inventarisatie:
                ShowConversationDetail<InventarisatieMessageType>(result.Conversation.Reference,
                    result.Scenario.Manager.ShvCertificate);
                break;
            case FlowType.BrEnHf:
                ShowConversationDetail<BrEnHfMessageType>(result.Conversation.Reference,
                    result.Scenario.Manager.ShvCertificate);
                break;
            case FlowType.Bewindvoering:
                ShowConversationDetail<BewindvoeringMessageType>(result.Conversation.Reference,
                    result.Scenario.Manager.ShvCertificate);
                break;
            case FlowType.PauzeSignaal3:
                ShowConversationDetail<PauzeSignaal3MessageType>(result.Conversation.Reference,
                    result.Scenario.Manager.ShvCertificate);
                break;
            case FlowType.ShvSignaal:
                ShowConversationDetail<ShvSignaalMessageType>(result.Conversation.Reference,
                    result.Scenario.Manager.ShvCertificate);
                break;
            default:
                throw new ArgumentOutOfRangeException(
                    $"Scenario conversation can't be displayed for {SelectedScenarioFlowType}");
        }
    }

    private void DgvScenarioResult_SelectionChanged(object sender, EventArgs e)
    {
        if (!(sender is DataGridView dgv) || dgv.SelectedRows.Count == 0)
        {
            return;
        }

        var group = dgv.SelectedRows[0].Tag as ScenarioResultGroup;
        if (group?.Results == null)
        {
            return;
        }

        BindData(dgvScenarioConverstations, group.Results, result => new object[]
        {
            result.Name,
            result.Conversation?.Reference,
            result.Success,
            "Open"
        });
    }

    private void TxtLogKey_TextChanged(object sender, EventArgs e)
    {
        if (_urlInfo != null)
        {
            _urlInfo.LogKey = LogKey;
        }
    }

    private object[] GetMessageBoxRow(MessageboxBerichtModel message)
    {
        return new object[]
        {
            message.Referentienummer,
            message.ConversatieReferentienummer,
            message.CrossReferentienummer,
            message.Berichttype,
            message.Vervallen,
            !message.Afzender.HasValue ? "SHV-Knooppunt" : message.Afzender.ToString(),
            message.Flow,
            message.Versie,
            message.Bijlagereferentienummers.Any() ? "Download" : null
        };
    }

    private async void BtnAutoReply_Click(object sender, EventArgs e)
    {
        try
        {
            SetAutoReplyEnabled(false);
            var certificateItem1 = cbAutoReplyCertificate1.SelectedItem as CertificateItem;
            var certificateItem2 = cbAutoReplyCertificate2.SelectedItem as CertificateItem;
            // TODO: support multiple flows
            var scenario = new AutoReplyScenario();

            var certificate = certificateItem1?.Certificate ?? certificateItem2?.Certificate;

            if (certificate == null)
            {
                AddMessage("No certificate selected");
                return;
            }

            AddMessage("Fetching the address book");
            var proxy = _urlInfo.GetAddressProxy(certificate, CurrentEnvironment, CurrentSystemVersion);
            if (proxy == null)
            {
                AddMessage("Something went wrong with creating the proxy.", false);
                return;
            }

            var organizations = await proxy.GetOrganizationsAsync(new AdresboekZoekBerichtModel());
            var messages = new List<MessageboxBerichtModel>();

            scenario.OnMessageReceived = message =>
            {
                messages.Add(message);
                dgvAutoReplyMessages.Rows.Insert(0, GetMessageBoxRow(message));
            };

            var certificates = new List<CertificateItem>();
            if (certificateItem1?.Certificate != null)
            {
                certificates.Add(certificateItem1);
            }

            if (certificateItem2?.Certificate != null)
            {
                certificates.Add(certificateItem2);
            }

            try
            {
                AutoReplyManager = new ScenarioManager(
                    scenario,
                    certificates,
                    organizations,
                    ResourceItemManager,
                    _urlInfo,
                    AddMessage,
                    CurrentSystemVersion,
                    CurrentEnvironment);

                AddMessage("Listening for messages");
                await scenario.ExecuteScenarioAsync();
            }
            catch (Exception scenarioEx)
            {
                AddMessage($"Scenario failed: {scenarioEx.Message}", false);
                throw;
            }
        }
        catch (Exception ex)
        {
            AddMessage($"Exception occured: {ex.Message}", false);
        }

        SetAutoReplyEnabled(true);
    }

    private void SetAutoReplyEnabled(bool canStart)
    {
        btnStartAutoReply.Enabled = canStart;
        btnStopAutoReply.Enabled = !canStart;
        cbAutoReplyCertificate1.Enabled = canStart;
        cbAutoReplyCertificate2.Enabled = canStart;
    }

    private void cbConversationFlowType_SelectedIndexChanged(object sender, EventArgs e)
    {
        var flowType = (FlowType)cbConversationFlowType.SelectedItem;

        SelectedConversationFlowType = flowType;

        var supportedVersions = FlowDefinitions.MajorVersionForFlowType(flowType);

        if (supportedVersions != null)
        {
            BindVersions(cbConversationFlowVersion, supportedVersions);
        }
    }

    private void cbLogbookFlowType_SelectedIndexChanged(object sender, EventArgs e)
    {
        var flowType = (FlowType)cbLogbookFlowType.SelectedItem;

        SelectedLogbookFlowType = flowType;

        switch (flowType)
        {
            case FlowType.Schuldbemiddeling:
                BindMessageTypes<SchuldbemiddelingMessageType>(cbLogMessageType);
                break;
            case FlowType.SbEnSk:
                BindMessageTypes<SbEnSkMessageType>(cbLogMessageType);
                break;
            case FlowType.Waarborgfonds:
                BindMessageTypes<WaarborgfondsMessageType>(cbLogMessageType);
                break;
            case FlowType.Inventarisatie:
                BindMessageTypes<InventarisatieMessageType>(cbLogMessageType);
                break;
            case FlowType.BrEnHf:
                BindMessageTypes<BrEnHfMessageType>(cbLogMessageType);
                break;
            case FlowType.Bewindvoering:
                BindMessageTypes<BewindvoeringMessageType>(cbLogMessageType);
                break;
            case FlowType.PauzeSignaal3:
                BindMessageTypes<PauzeSignaal3MessageType>(cbLogMessageType);
                break;
            case FlowType.ShvSignaal:
                BindMessageTypes<ShvSignaalMessageType>(cbLogMessageType);
                break;
            default:
                throw new NotImplementedException($"No message types available for {flowType}");
        }

        var supportedVersions = FlowDefinitions.VersionsForFlowType(flowType);

        if (supportedVersions != null)
        {
            BindVersions(cbLogbookFlowVersion, supportedVersions);
        }
    }

    private void shvBrowseButton_Click(object sender, EventArgs e)
    {
        SetAttachments(ShvAddedAttachments, AddMessage);

        shvDeleteAttachmentButton.Visible = true;
    }

    private void seBrowseButton_Click(object sender, EventArgs e)
    {
        SetAttachments(SeAddedAttachments, AddMessage);

        seClearAttachmentButton.Visible = true;
    }

    private void seClearAttachmentButton_Click(object sender, EventArgs e)
    {
        SeAddedAttachments.Clear();
        seClearAttachmentButton.Visible = false;
    }

    private void dgvMessageBoxMessages_CellContentClick(object sender, DataGridViewCellEventArgs e)
    {
        if (!(dgvMessageBoxMessages.Columns[e.ColumnIndex] is DataGridViewButtonColumn) || e.RowIndex == -1)
        {
            return;
        }

        var item = GetSelectedItem<MessageboxBerichtModel>(dgvMessageBoxMessages);

        if (item.Bijlagereferentienummers.Count == 0)
        {
            AddMessage("No attachments found for the selected message");
            return;
        }

        var certificate = GetCertificate(cbMessageBoxCertificate);

        var messageProxy = _urlInfo.GetMessageBoxProxy(certificate, CurrentEnvironment, CurrentSystemVersion);

        var attachmentForm = new DownloadAttachmentsForm(item, messageProxy);

        try
        {
            attachmentForm.ShowDialog();
        }
        catch (InvalidOperationException)
        {
            //Something went wrong while trying to download the attachment. We catch the exception else the application exits forcefully. 
            attachmentForm.Close();
            AddMessage("Something went wrong while trying to download the attachment");
        }
    }

    private void shvDeleteAttachmentButton_Click(object sender, EventArgs e)
    {
        ShvAddedAttachments.Clear();
        shvDeleteAttachmentButton.Visible = false;
    }

    private void FlowListBox_DrawItem(object sender, DrawItemEventArgs e)
    {
        var f = e.Font;
        if ((e.State & DrawItemState.Selected) == DrawItemState.Selected)
        {
            f = new Font(e.Font, FontStyle.Bold);
        }

        e.DrawBackground();
        e.Graphics.FillRectangle(new SolidBrush(SystemColors.Control), e.Bounds);

        e.Graphics.DrawString(((ListBox)sender).Items[e.Index].ToString(), f, new SolidBrush(Color.DimGray),
            e.Bounds);
        e.DrawFocusRectangle();
    }

    private void cbCurrentEnvironment_SelectedIndexChanged(object sender, EventArgs e)
    {
        Text = $@"S2S Test Application - Schuldenknooppunt - Omgeving: {CurrentEnvironment.Name}";
    }

    #region Eventhandlers

    private void BtnStopAutoReply_Click(object sender, EventArgs e)
    {
        AutoReplyManager?.Stop();

        SetAutoReplyEnabled(true);
    }

    private void cbShvFlowType_SelectedIndexChanged(object sender, EventArgs e)
    {
        var flowType = (FlowType)cbShvFlowType.SelectedItem;

        SelectedShvFlowType = flowType;

        var supportedVersions = FlowDefinitions.VersionsForFlowType(flowType);

        if (supportedVersions != null)
        {
            BindVersions(cbShvMessageVersion, supportedVersions);
        }
    }

    private void CbShvMessageVersion_SelectedIndexChanged(object sender, EventArgs e)
    {
        switch (SelectedShvFlowType)
        {
            case FlowType.Schuldbemiddeling:
                var schuldbemiddelingItems = ResourceItemManager
                    .GetAllResources<SchuldbemiddelingMessageType>(SelectedShvFlowType ?? DefaultFlowType,
                        cbShvMessageVersion.SelectedItem.ToString(), false).ToArray();
                BindMessageTypes(cbShvMessageType, schuldbemiddelingItems);
                break;
            case FlowType.SbEnSk:
                var sbEnSkItems = ResourceItemManager
                    .GetAllResources<SbEnSkMessageType>(SelectedShvFlowType ?? DefaultFlowType,
                        cbShvMessageVersion.SelectedItem.ToString(), false).ToArray();
                BindMessageTypes(cbShvMessageType, sbEnSkItems);
                break;
            case FlowType.Waarborgfonds:
                var waarborgfondsItems = ResourceItemManager
                    .GetAllResources<WaarborgfondsMessageType>(SelectedShvFlowType ?? DefaultFlowType,
                        cbShvMessageVersion.SelectedItem.ToString(), false).ToArray();
                BindMessageTypes(cbShvMessageType, waarborgfondsItems);
                break;
            case FlowType.Inventarisatie:
                var inventarisatieItems = ResourceItemManager
                    .GetAllResources<InventarisatieMessageType>(SelectedShvFlowType ?? DefaultFlowType,
                        cbShvMessageVersion.SelectedItem.ToString(), false).ToArray();
                BindMessageTypes(cbShvMessageType, inventarisatieItems);
                break;
            case FlowType.BrEnHf:
                var brEnHfItems = ResourceItemManager
                    .GetAllResources<BrEnHfMessageType>(SelectedShvFlowType ?? DefaultFlowType,
                        cbShvMessageVersion.SelectedItem.ToString(), false).ToArray();
                BindMessageTypes(cbShvMessageType, brEnHfItems);
                break;
            case FlowType.Bewindvoering:
                var bewindvoerderItems = ResourceItemManager
                    .GetAllResources<BewindvoeringMessageType>(SelectedShvFlowType ?? DefaultFlowType,
                        cbShvMessageVersion.SelectedItem.ToString(), false).ToArray();
                BindMessageTypes(cbShvMessageType, bewindvoerderItems);
                break;
            case FlowType.PauzeSignaal3:
                var pauzeSignaal3Items = ResourceItemManager
                    .GetAllResources<PauzeSignaal3MessageType>(SelectedShvFlowType ?? DefaultFlowType,
                        cbShvMessageVersion.SelectedItem.ToString(), false).ToArray();
                BindMessageTypes(cbShvMessageType, pauzeSignaal3Items);
                break;
            case FlowType.ShvSignaal:
                var shvSignaalItems = ResourceItemManager
                    .GetAllResources<ShvSignaalMessageType>(SelectedShvFlowType ?? DefaultFlowType,
                        cbShvMessageVersion.SelectedItem.ToString(), false).ToArray();
                BindMessageTypes(cbShvMessageType, shvSignaalItems);
                break;
            default:
                throw new InvalidOperationException($"No Resource items available for {SelectedShvFlowType}");
        }
    }

    private void CbShvMessageType_SelectedIndexChanged(object sender, EventArgs e)
    {
        switch (SelectedShvFlowType)
        {
            case FlowType.Schuldbemiddeling:
                var schuldbemiddelingItem =
                    GetSelectedValue<ResourceItem<SchuldbemiddelingMessageType>>(cbShvMessageType);
                SelectedShvSchuldbemiddelingResourceItem = schuldbemiddelingItem;
                UpdateTextBoxContent(rtbShvMessageContent, schuldbemiddelingItem);
                break;
            case FlowType.SbEnSk:
                var sbEnSkItem = GetSelectedValue<ResourceItem<SbEnSkMessageType>>(cbShvMessageType);
                SelectedShvSbEnSkResourceItem = sbEnSkItem;
                UpdateTextBoxContent(rtbShvMessageContent, sbEnSkItem);
                break;
            case FlowType.Waarborgfonds:
                var waarborgfondsItem = GetSelectedValue<ResourceItem<WaarborgfondsMessageType>>(cbShvMessageType);
                SelectedShvWaarborgfondsResourceItem = waarborgfondsItem;
                UpdateTextBoxContent(rtbShvMessageContent, waarborgfondsItem);
                break;
            case FlowType.Inventarisatie:
                var inventariesatieItem =
                    GetSelectedValue<ResourceItem<InventarisatieMessageType>>(cbShvMessageType);
                SelectedShvInventarisatieResourceItem = inventariesatieItem;
                UpdateTextBoxContent(rtbShvMessageContent, inventariesatieItem);
                break;
            case FlowType.BrEnHf:
                var brEnHfItems = GetSelectedValue<ResourceItem<BrEnHfMessageType>>(cbShvMessageType);
                SelectedShvBrEnHfResourceItem = brEnHfItems;
                UpdateTextBoxContent(rtbShvMessageContent, brEnHfItems);
                break;
            case FlowType.Bewindvoering:
                var bewindvoeringItems = GetSelectedValue<ResourceItem<BewindvoeringMessageType>>(cbShvMessageType);
                SelectedShvBewindvoeringResourceItem = bewindvoeringItems;
                UpdateTextBoxContent(rtbShvMessageContent, bewindvoeringItems);
                break;
            case FlowType.PauzeSignaal3:
                var pauzeSignaal3Items = GetSelectedValue<ResourceItem<PauzeSignaal3MessageType>>(cbShvMessageType);
                SelectedShvPauzeSignaal3ResourceItem = pauzeSignaal3Items;
                UpdateTextBoxContent(rtbShvMessageContent, pauzeSignaal3Items);
                break;
            case FlowType.ShvSignaal:
                var shvSignaalItems = GetSelectedValue<ResourceItem<ShvSignaalMessageType>>(cbShvMessageType);
                SelectedShvSignaalResourceItem = shvSignaalItems;
                UpdateTextBoxContent(rtbShvMessageContent, shvSignaalItems);
                break;
            default:
                throw new ArgumentOutOfRangeException($"Message content can't be loaded for {SelectedShvFlowType}");
        }
    }

    private void cbSeFlowType_SelectedIndexChanged(object sender, EventArgs e)
    {
        var flowType = (FlowType)cbSeFlowType.SelectedItem;

        SelectedSeFlowType = flowType;

        var supportedVersions = FlowDefinitions.VersionsForFlowType(flowType);

        if (supportedVersions != null)
        {
            BindVersions(cbSeMessageVersion, supportedVersions);
        }
    }

    private void CbSeMessageVersion_SelectedIndexChanged(object sender, EventArgs e)
    {
        switch (SelectedSeFlowType)
        {
            case FlowType.Schuldbemiddeling:
                var schuldbemiddelingItems = ResourceItemManager
                    .GetAllResources<SchuldbemiddelingMessageType>(SelectedSeFlowType ?? DefaultFlowType,
                        cbSeMessageVersion.SelectedItem.ToString(), true).ToArray();
                BindMessageTypes(cbSeMessageType, schuldbemiddelingItems);
                break;
            case FlowType.SbEnSk:
                var sbEnSkItems = ResourceItemManager
                    .GetAllResources<SbEnSkMessageType>(SelectedSeFlowType ?? DefaultFlowType,
                        cbSeMessageVersion.SelectedItem.ToString(), true).ToArray();
                BindMessageTypes(cbSeMessageType, sbEnSkItems);
                break;
            case FlowType.Waarborgfonds:
                var waarborgfondsItems = ResourceItemManager
                    .GetAllResources<WaarborgfondsMessageType>(SelectedSeFlowType ?? DefaultFlowType,
                        cbSeMessageVersion.SelectedItem.ToString(), true).ToArray();
                BindMessageTypes(cbSeMessageType, waarborgfondsItems);
                break;
            case FlowType.Inventarisatie:
                var inventarisatieItems = ResourceItemManager
                    .GetAllResources<InventarisatieMessageType>(SelectedSeFlowType ?? DefaultFlowType,
                        cbSeMessageVersion.SelectedItem.ToString(), true).ToArray();
                BindMessageTypes(cbSeMessageType, inventarisatieItems);
                break;
            case FlowType.BrEnHf:
                var brEnHfItems = ResourceItemManager
                    .GetAllResources<BrEnHfMessageType>(SelectedSeFlowType ?? DefaultFlowType,
                        cbSeMessageVersion.SelectedItem.ToString(), true).ToArray();
                BindMessageTypes(cbSeMessageType, brEnHfItems);
                break;
            case FlowType.Bewindvoering:
                var bewindvoeringItems = ResourceItemManager
                    .GetAllResources<BewindvoeringMessageType>(SelectedSeFlowType ?? DefaultFlowType,
                        cbSeMessageVersion.SelectedItem.ToString(), true).ToArray();
                BindMessageTypes(cbSeMessageType, bewindvoeringItems);
                break;
            case FlowType.PauzeSignaal3:
                var pauzeSignaal3Items = ResourceItemManager
                    .GetAllResources<PauzeSignaal3MessageType>(SelectedSeFlowType ?? DefaultFlowType,
                        cbSeMessageVersion.SelectedItem.ToString(), true).ToArray();
                BindMessageTypes(cbSeMessageType, pauzeSignaal3Items);
                break;
            case FlowType.ShvSignaal:
                BindMessageTypes(cbSeMessageType, Array.Empty<ResourceItem<ShvSignaalMessageType>>());
                break;
            default:
                throw new InvalidOperationException($"No Resource items available for {SelectedSeFlowType}");
        }
    }

    private void CbSeMessageType_SelectedIndexChanged(object sender, EventArgs e)
    {
        switch (SelectedSeFlowType)
        {
            case FlowType.Schuldbemiddeling:
                var schuldbemiddelingItem =
                    GetSelectedValue<ResourceItem<SchuldbemiddelingMessageType>>(cbSeMessageType);
                SelectedSeSchuldbemiddelingResourceItem = schuldbemiddelingItem;
                UpdateTextBoxContent(rtbSeContent, schuldbemiddelingItem);
                break;
            case FlowType.SbEnSk:
                var sbEnSkItem = GetSelectedValue<ResourceItem<SbEnSkMessageType>>(cbSeMessageType);
                SelectedSeSbEnSkResourceItem = sbEnSkItem;
                UpdateTextBoxContent(rtbSeContent, sbEnSkItem);
                break;
            case FlowType.Waarborgfonds:
                var waarborgfondsItem = GetSelectedValue<ResourceItem<WaarborgfondsMessageType>>(cbSeMessageType);
                SelectedSeWaarborgfondsResourceItem = waarborgfondsItem;
                UpdateTextBoxContent(rtbSeContent, waarborgfondsItem);
                break;
            case FlowType.Inventarisatie:
                var inventarisatieItem = GetSelectedValue<ResourceItem<InventarisatieMessageType>>(cbSeMessageType);
                SelectedSeInventarisatieResourceItem = inventarisatieItem;
                UpdateTextBoxContent(rtbSeContent, inventarisatieItem);
                break;
            case FlowType.BrEnHf:
                var brEnHfItems = GetSelectedValue<ResourceItem<BrEnHfMessageType>>(cbSeMessageType);
                SelectedSeBrEnHfResourceItem = brEnHfItems;
                UpdateTextBoxContent(rtbSeContent, brEnHfItems);
                break;
            case FlowType.Bewindvoering:
                var bewindvoeringItems = GetSelectedValue<ResourceItem<BewindvoeringMessageType>>(cbSeMessageType);
                SelectedSeBewindvoeringResourceItem = bewindvoeringItems;
                UpdateTextBoxContent(rtbSeContent, bewindvoeringItems);
                break;
            case FlowType.PauzeSignaal3:
                var pauzeSignaal3Items = GetSelectedValue<ResourceItem<PauzeSignaal3MessageType>>(cbSeMessageType);
                SelectedSePauzeSignaal3ResourceItem = pauzeSignaal3Items;
                UpdateTextBoxContent(rtbSeContent, pauzeSignaal3Items);
                break;
            default:
                throw new ArgumentOutOfRangeException($"Message content can't be loaded for {SelectedSeFlowType}");
        }
    }

    private void cbCurrentVersion_SelectedIndexChanged(object sender, EventArgs e)
    {
        BindOrganizationTypes(cbOrganisationType, CurrentSystemVersion);
        BindOrganizationTypes(cbLogSenderOrganizationType, CurrentSystemVersion);
        BindOrganizationTypes(cbLogReceiverOrganizationType, CurrentSystemVersion);
    }

    private async void Button2_Click(object sender, EventArgs e)
    {
        await ConfirmMessagesAsync(GetCertificate(cbMessageBoxCertificate), txtMessageBoxBatchId.Text);
    }

    private void BtnNewConversationRef_Click(object sender, EventArgs e)
    {
        SetGuid(txtConversationRef);
    }

    private void BtnNewCrossRef_Click(object sender, EventArgs e)
    {
        SetGuid(txtCrossReference);
    }

    private void BtnNewRef_Click(object sender, EventArgs e)
    {
        SetGuid(txtReference);
    }

    private async void Button3_Click(object sender, EventArgs e)
    {
        await LoadOrganisationsAsync();
    }

    private async void Button1_Click(object sender, EventArgs e)
    {
        lblMessageBoxBatchValidTo.ResetText();
        txtMessageBoxBatchId.Clear();
        rtbMessageBoxMessageContent.Clear();

        await GetMessagesAsync(GetCertificate(cbMessageBoxCertificate));
    }

    public void AddMessage(string message, bool success = true)
    {
        AddColoredMessage(richTextBox1, message, success);
    }

    private void DataGridView1_Click(object sender, EventArgs e)
    {
        var item = GetSelectedItem<MessageboxBerichtModel>(dgvMessageBoxMessages);
        if (item == null)
        {
            return;
        }

        rtbMessageBoxMessageContent.Text = item.Content ?? string.Empty;
        RtbHelper.HighlightRTF(rtbMessageBoxMessageContent);
    }

    private async void BtnShvSend_Click(object sender, EventArgs e)
    {
        switch (SelectedShvFlowType)
        {
            case FlowType.Schuldbemiddeling:
                await ExecuteShvActionAsync(SelectedShvSchuldbemiddelingResourceItem);
                break;
            case FlowType.SbEnSk:
                await ExecuteShvActionAsync(SelectedShvSbEnSkResourceItem);
                break;
            case FlowType.Waarborgfonds:
                await ExecuteShvActionAsync(SelectedShvWaarborgfondsResourceItem);
                break;
            case FlowType.Inventarisatie:
                await ExecuteShvActionAsync(SelectedShvInventarisatieResourceItem);
                break;
            case FlowType.BrEnHf:
                await ExecuteShvActionAsync(SelectedShvBrEnHfResourceItem);
                break;
            case FlowType.Bewindvoering:
                await ExecuteShvActionAsync(SelectedShvBewindvoeringResourceItem);
                break;
            case FlowType.PauzeSignaal3:
                await ExecuteShvActionAsync(SelectedShvPauzeSignaal3ResourceItem);
                break;
            case FlowType.ShvSignaal:
                await ExecuteShvActionAsync(SelectedShvSignaalResourceItem);
                break;
            case null:
                AddMessage("Flow niet geselecteerd", false);
                break;
            default:
                throw new ArgumentOutOfRangeException();
        }
    }

    private async void BtnSeSend_Click(object sender, EventArgs e)
    {
        switch (SelectedSeFlowType)
        {
            case FlowType.Schuldbemiddeling:
                await ExecuteSeActionAsync(SelectedSeSchuldbemiddelingResourceItem);
                break;
            case FlowType.SbEnSk:
                await ExecuteSeActionAsync(SelectedSeSbEnSkResourceItem);
                break;
            case FlowType.Waarborgfonds:
                await ExecuteSeActionAsync(SelectedSeWaarborgfondsResourceItem,
                    GetBerichtHeaderWaarborgfonds(SelectedSeWaarborgfondsResourceItem.Berichttype));
                break;
            case FlowType.Inventarisatie:
                await ExecuteSeActionAsync(SelectedSeInventarisatieResourceItem);
                break;
            case FlowType.BrEnHf:
                await ExecuteSeActionAsync(SelectedSeBrEnHfResourceItem);
                break;
            case FlowType.Bewindvoering:
                await ExecuteSeActionAsync(SelectedSeBewindvoeringResourceItem);
                break;
            case FlowType.PauzeSignaal3:
                await ExecuteSeActionAsync(SelectedSePauzeSignaal3ResourceItem);
                break;
            case FlowType.ShvSignaal:
                AddMessage("Flow niet ondersteunt i.c.m. dit organisatie type", false);
                break;
            case null:
                AddMessage("Flow niet geselecteerd", false);
                break;
            default:
                throw new ArgumentOutOfRangeException();
        }
    }

    private static void UpdateTextBoxContent<TMessageType>(RichTextBox textBox,
        ResourceItem<TMessageType> resourceItem)
        where TMessageType : Enum
    {
        if (resourceItem == null)
        {
            return;
        }

        textBox.Text = resourceItem.Content;
        RtbHelper.HighlightRTF(textBox);
    }

    private void CbMessageBoxCertificate_SelectedIndexChanged(object sender, EventArgs e)
    {
        dgvMessageBoxMessages.Rows.Clear();
        txtMessageBoxBatchId.Clear();
    }

    private void BtnSeGenerateReference_Click(object sender, EventArgs e)
    {
        SetGuid(txtSeReference);
    }

    private void BtnShvClearCrossReference_Click(object sender, EventArgs e)
    {
        txtCrossReference.Clear();
    }

    private void BtnSeClearCrossReference_Click(object sender, EventArgs e)
    {
        txtSeCrossReference.Clear();
    }

    private void BtnSeClearConversationReference_Click(object sender, EventArgs e)
    {
        txtSeConversationReference.Clear();
    }

    private async void BtnLoadLogbook_Click(object sender, EventArgs e)
    {
        switch (SelectedLogbookFlowType)
        {
            case FlowType.Schuldbemiddeling:
                await LoadLogbookAsync<SchuldbemiddelingMessageType>();
                break;
            case FlowType.SbEnSk:
                await LoadLogbookAsync<SbEnSkMessageType>();
                break;
            case FlowType.Waarborgfonds:
                await LoadLogbookAsync<WaarborgfondsMessageType>();
                break;
            case FlowType.Inventarisatie:
                await LoadLogbookAsync<InventarisatieMessageType>();
                break;
            case FlowType.BrEnHf:
                await LoadLogbookAsync<BrEnHfMessageType>();
                break;
            case FlowType.Bewindvoering:
                await LoadLogbookAsync<BewindvoeringMessageType>();
                break;
            case FlowType.PauzeSignaal3:
                await LoadLogbookAsync<PauzeSignaal3MessageType>();
                break;
            case FlowType.ShvSignaal:
                await LoadLogbookAsync<ShvSignaalMessageType>();
                break;
            default:
                throw new ArgumentOutOfRangeException($"Logbook not loadable for {SelectedLogbookFlowType}");
        }
    }

    private async void BtnOrgSaveChanges_Click(object sender, EventArgs e)
    {
        await UpdateOrganizationAsync();
    }

    private async void BtnOrgLoad_Click(object sender, EventArgs e)
    {
        await LoadOrganizationDetailAsync();
    }

    private void cbScenarioFlowType_SelectedIndexChanged(object sender, EventArgs e)
    {
        var flowType = (FlowType)cbScenarioFlowType.SelectedItem;

        SelectedScenarioFlowType = flowType;

        LoadScenarios(flowType);
    }

    #endregion
}