﻿using System.Net;
using System.Reflection;
using System.Xml;

namespace Infrastructure.Utils.Xml;

public class ResourceXmlResolver : XmlResolver
{
    // We use a fake uri path to be able to easily perform the xml resolution
    private readonly Uri _fakePath = new("\\\\fake");

    public ResourceXmlResolver(Assembly assembly)
    {
        Assembly = assembly;
    }

    private Assembly Assembly { get; }
    public string Prefix { private get; set; }

    public override ICredentials Credentials
    {
        set => throw new NotImplementedException();
    }

    public override object GetEntity(Uri absoluteUri, string role, Type ofObjectToReturn)
    {
        var resourceName = _fakePath.MakeRelativeUri(absoluteUri)
            .ToString()
            .Replace('/', '.');

        var result = Assembly.GetManifestResourceStream(resourceName);

        return result;
    }

    public override Uri ResolveUri(Uri baseUri, string relativeUri)
    {
        if (baseUri.IsAbsoluteUri)
        {
            return new Uri(baseUri, relativeUri);
        }

        return new Uri(_fakePath, Path.Combine(Prefix.Replace('.', '/'), relativeUri));
    }
}