﻿using System.Text.RegularExpressions;
using SoapEndpoint.Service.Proxy.Infrastructure;
using SoapEndpoint.Service.Proxy.Models.Flows;

namespace SoapEndpoint.Service.Proxy.ResourceItems;

public class ResourceItem<TMessageType> where TMessageType : Enum
{
    public ResourceItem(ResourceManagerModel resourceManagerModel, TMessageType messageType, string? typeName = null)
    {
        TypeName = string.IsNullOrEmpty(typeName) ? messageType.ToString() : typeName;

        ResourceManagerModel =
            resourceManagerModel ?? throw new ArgumentNullException(nameof(resourceManagerModel));
        FlowType = resourceManagerModel.FlowType;
        Berichttype = messageType;
        Content = resourceManagerModel.ResourceManager.GetString(TypeName) ?? string.Empty;

        var strippedMessageType = TypeName.Replace("SHV_", "").Replace("SE_", "");
        Name = Regex.Replace(strippedMessageType, "(\\B[A-Z])", " $1");
        SeType = TypeName.StartsWith("SE_");
    }

    public ResourceManagerModel ResourceManagerModel { get; }
    public FlowType FlowType { get; }
    public string Content { get; }
    public TMessageType Berichttype { get; }
    public string TypeName { get; }
    public string Name { get; }
    public bool SeType { get; }
}